from scipy.stats import norm
import pandas as pd
import numpy as np


def PAA_step(df_INT, w):
    """
    Piecewise Aggregate Approximation (PAA) (Keogh et. al., 2001) for multivariate 
    time series.

    Inputs:
    - df_INT = Resulting DataFrame from the normalization step 
    - w = number of segments
    """

    n = len(df_INT)

    # Resample dataframe
    new_index = df_INT.index.repeat(w)
    df_temp = df_INT.iloc[:, 1:].reindex(
        new_index)  # Do not apply on the time column

    # Create new re-indexed dataframe from previously re-indexed dataframe
    df_PAA = pd.DataFrame(data=df_temp.to_numpy(),
                          index=[i for i in range(w) for k in range(n)],
                          columns=df_INT.columns[1:])

    # Average segments
    df_PAA = df_PAA.groupby(df_PAA.index).mean()

    # Create a new correct time column
    new_t = pd.date_range(start=df_INT['t'].iloc[0],
                          end=df_INT['t'].iloc[-1],
                          periods=w+1)
    df_PAA['t'] = new_t[:-1]

    # Reorder columns
    cols = df_PAA.columns.tolist()
    new_cols = cols[-1:] + cols[:-1]
    df_PAA = df_PAA[new_cols]

    return df_PAA


def SAX_step(df_INT, df_PAA, a):
    """
    Symbolic representation of the normalized DataFrame.

    Inputs:
    - df_PAA = PAA of a pandas dataframe
    - a = alphabet size.
    """

    # Define alphabet from alphabet size and corresponding breakpoints
    alphabet = np.array([chr(i) for i in range(97, 97 + a)])
    breakpoints = norm.ppf(np.linspace(0, 1, a+1)[1:-1])

    # Initialize output DataFrame
    df_SAX = pd.DataFrame(index=df_PAA.index)
    df_SAX['t'] = df_PAA['t']

    # Process each column of the input dataframe
    for column in df_INT.columns[1:]:

        # Convert each value into a letter
        n_timestamps = len(df_PAA)
        discretization = []
        for t in range(n_timestamps):

            discretization.append(alphabet[np.searchsorted(breakpoints,
                                                           df_PAA[column].iloc[t],
                                                           side='left')])

        df_SAX[column] = discretization

    return df_SAX, alphabet[-1], alphabet[0]
